//------------------------------------------------------------------------------
// ObjectAsIso.inc
//
// An include file which works with the pattern object capability of POVRay
// to treat objects as if they have some field strength - so they work as
// isosurface functions.  This technique is expensive.  The Object.pov
// render times are on the order of 12.5 minutes elapsed on a slightly
// overclocked i7 920 quad core processor (100min real). A patch of the
// POVRay source runs in 2 minutes (17min real).
//
// For something as simple as the Object.pov CSG object it would be
// much faster to use equations or, Christoph Hormann and Wlodzimierz ABX
// Skiba's IsoCSG Library (iso_csg.inc) to get an equivalent isosurface
// object.  That said, there may be situations where the object is complex
// and this becomes the only viable means to an equivalent isosurface.  For
// example, I've run "Lucy," from The Stanford 3D Sanning Repository (via
// Robert McGregor), as an isosurface.
//
// Typical usage:
//
//    #include "ObjectAsIso.inc"
//    #declare Thing = difference { box {-1,1} sphere {0,1.2} }
//    object {
//      isosurface{
//        function{0.100-ObjectAsIso(Thing,0.05,20)+(f_noise3d(30*x,30*y,30*z)/5) }
//        ...
//
// Arguments:
//    arg1 - The object
//    arg2 - The strength of the field. The object surface will be at
//           a strength of approximately 0.0 ramping up to 1.0 at this
//           distance inside the original CSG surface.
//    arg3 - The depth of recursion during the binary search for the field
//           strength at each point tested.
//
// Usage notes:
//    1) This method uses a set of evenly spaced sample points on
//       the surface of a sphere with a varying radius. The maximum radius of
//       arg2 is used during the initial, do we test further test. Features
//       smaller than arg2 can be missed if not attached/near a feature >= arg2.
//       Pick the arg2 field strength to fit your object. Further, the
//       'main' of the object should have a thickness of > arg2. Generally,
//       the larger this value, the lower the max gradient. A value of 0.05
//       has worked reasonably well for most objects tried thus far.
//    2) Unsure why, but it is best to build or scale the object down
//       near the origin around the default isosurface container size of a
//       -1,1 box. Isosurfaces run faster there and being this method is so
//       slow, it often makes a huge difference in run times.
//    3) The search depth can be adjusted by arg3. Less depth runs faster, but
//       results in larger max gradients. Thus far, a depth of 20 seems a
//       decent tradeoff.
//    4) There is probably no reason to turn objects into isosurfaces if just
//       rendering the original object, but, DO add the noise! With this method
//       the additional noise, or pattern almost always helps with the run time
//       where normally adding noise adds run time. Further, we are sampling points,
//       which means artifacts of the sampling are more apparent on the plain/smooth
//       object-as-isosurface.
//    5) There is currently no rounding/softening of sharp edges. These can
//       appear ragged in the image. Use objects with rounded edges unless
//       a ragged appearence is the goal. For of rounded object package, see:
//       http://lib.povray.org/collection/roundedge/cousin ricky 1.0/roundedge.html
//    6) In a typical set up it would make sense to develop a scene using the 'fast'
//       object rendering and then switch to the object-as-isosurface for the final
//       render.
//               ...
//               #declare SimpleObjects = on;  // on | off
//               ...
//               #if (SimpleObjects)
//                    object { Lucy
//                             texture { pigment { color RGB(<1.0,0.78824,0.00392>) } }
//                           }
//               #else
//                    object {
//                         #declare LucyP = function{pattern{object {Lucy}}}
//                         isosurface{
//                           function{0.100-LucyP(x,y,z) }
//               ...
//               #end
//    7) Likely better to render large images and then scale down as opposed
//       to using AA. In addition to this method being extremely expensive per
//       ray cast, isosurfaces in general occasionally interact with the POVRay
//       SMP implementation in an unfortunate way. It can be, especially when the
//       image is smaller, that a compute intensive region of the isosurface
//       generated image ends up allocated to a single thread.
//
// Thoughts:
//    -  Adding some rounding/smoothing of objects with this method is likely
//       doable, but at what expense is not clear. The sampling returns a
//       proximity value, N/23 currently, but only three value states are used.
//       The implmentation and expense for a more complex field calculation is
//       not clear. The sampling method on its own should make for a decent
//       proximity pattern, but no work has been done here.
//
// Credits:
//    - Paul Bourke, for c code to generate equidistant sample points on a sphere.
//    - Mike Williams, for the isosurface tutorial which can be found at:
//      http://www.econym.demon.co.uk/isotut/index.htm, and also for a recent
//      mention of the 'select' function in answering someones question.
//    - Thorsten Froehlich et al, for the VM implementation. The SDL include, which
//      was written first, is not that much slower than the POVRay code patch!
//    - Robert McGregor, for providing mesh2 encoding of Stanford's 'Lucy' mesh.
//    - Bill Pragnell, for providing the mesh2 archimedean_solids.inc which
//      can be found at: http://lib.povray.org/searchcollection/index.php
//    - Intel's i7 920.
//
//
// Version 1.0                                                   WFP 09/14/09
//------------------------------------------------------------------------------

#include "functions.inc"

#macro ObjectAsIso(_Object,_Sorg,_Depth)

   #local ObjectAsIsoPattern = function{pattern{object {_Object}}}

   #local ObjectAsIsoProx = function (x,y,z,S) { (
            ObjectAsIsoPattern( x,              y,               z              ) +
            ObjectAsIsoPattern( x+(S*0.118898), y+(S*0.917706 ), z+(S*0.379049) ) +
            ObjectAsIsoPattern( x+(S*0.707714), y+(S*0.481318 ), z+(S*0.517179) ) +
            ObjectAsIsoPattern( x+(S*0.610501), y+(S*0.773853 ), z-(S*0.168644) ) +
            ObjectAsIsoPattern( x+(S*0.052713), y+(S*0.747854 ), z-(S*0.661767) ) +
            ObjectAsIsoPattern( x+(S*0.623082), y+(S*0.285055 ), z-(S*0.728363) ) +
            ObjectAsIsoPattern( x+(S*0.385597), y-(S*0.825183 ), z+(S*0.412781) ) +
            ObjectAsIsoPattern( x+(S*0.845586), y-(S*0.247858 ), z+(S*0.472811) ) +
            ObjectAsIsoPattern( x+(S*0.255258), y-(S*0.298976 ), z+(S*0.919487) ) +
            ObjectAsIsoPattern( x+(S*0.938207), y-(S*0.218780 ), z-(S*0.268146) ) +
            ObjectAsIsoPattern( x+(S*0.174757), y-(S*0.938110 ), z-(S*0.299015) ) +
            ObjectAsIsoPattern( x+(S*0.446882), y-(S*0.440226 ), z-(S*0.778780) ) +
            ObjectAsIsoPattern( x-(S*0.937512), y+(S*0.347669 ), z+(S*0.014056) ) +
            ObjectAsIsoPattern( x-(S*0.532628), y+(S*0.620243 ), z+(S*0.575852) ) +
            ObjectAsIsoPattern( x-(S*0.447006), y+(S*0.887248 ), z-(S*0.113917) ) +
            ObjectAsIsoPattern( x-(S*0.622939), y+(S*0.414059 ), z-(S*0.663704) ) +
            ObjectAsIsoPattern( x-(S*0.045097), y+(S*0.083557 ), z-(S*0.995482) ) +
            ObjectAsIsoPattern( x-(S*0.887956), y-(S*0.327793 ), z+(S*0.322623) ) +
            ObjectAsIsoPattern( x-(S*0.327528), y-(S*0.727050 ), z+(S*0.603427) ) +
            ObjectAsIsoPattern( x-(S*0.446895), y-(S*0.053668 ), z+(S*0.892975) ) +
            ObjectAsIsoPattern( x-(S*0.535803), y-(S*0.837654 ), z-(S*0.106071) ) +
            ObjectAsIsoPattern( x-(S*0.869977), y-(S*0.246679 ), z-(S*0.426953) ) +
            ObjectAsIsoPattern( x-(S*0.287314), y-(S*0.580587 ), z-(S*0.761820) )
          )
   }

   #local ObjectAsPatternBsearch = array[_Depth+1];

   #local Cur = 0;    // Below must eliminate run of very small constant values arg2 outside object to object surface
   #local ObjectAsPatternBsearch[Cur]= function (x,y,z,S,Sorg,Depth) { select(S-(2*(Sorg/pow(2,Depth-1))),0,S/Sorg,S/Sorg)
                                                                     }
   #local Cur = 1;
   #while (Cur <=_Depth)
     #local ObjectAsPatternBsearch[Cur]= function (x,y,z,S,Sorg,Depth) { select(23-ObjectAsIsoProx(x,y,z,S),0,
                                                                          ObjectAsPatternBsearch[Cur-1](x,y,z,S+(Sorg/pow(2,Depth+1)),Sorg,Depth+1),
                                                                          ObjectAsPatternBsearch[Cur-1](x,y,z,S-(Sorg/pow(2,Depth+1)),Sorg,Depth+1)
                                                                         )
                                                                       }
     #local Cur = Cur + 1;
   #end

   select(23-ObjectAsIsoProx(x,y,z,_Sorg),1,1,
    select(ObjectAsIsoProx(x,y,z,_Sorg),0,0,
     ObjectAsPatternBsearch[_Depth](x,y,z,_Sorg-(_Sorg/2),_Sorg,1)
    )
   )

#end

